package testkinect;

import javax.swing.*;
import javax.swing.event.*;
import java.awt.*;
import java.awt.image.*;
import java.awt.event.*;
import java.io.*;

import utildll.*;

public class TestKINECT104 extends JFrame implements Runnable, ActionListener {
	Wrp_KINECT wk = new Wrp_KINECT();
	Wrp_INuiSensor wins = null;
	Wrp_INuiCoordinateMapper wmap = null;
	Wrp_Kernel32 wkrnl = new Wrp_Kernel32();
	Wrp_JInvoker wjinv = new Wrp_JInvoker();
	long[] kinect = {0};
	long[] hColor = {0};
	long[] hDepth = {0};
	long[] mapper = {0};
	long hColorEvent = 0;
	long hDepthEvent = 0;
	long[] hEvents;
	int[] width = {0};
	int[] height = {0};
	int[] color;
	int[] depth; //NUI_DEPTH_IMAGE_PIXEL {USHORT playerIndex; USHORT depth;}
	int[] colorPoint; // NUI_COLOR_IMAGE_POINT {LONG x; LONG y;}
	
	// fusion data
	Wrp_KinectFusion wkf = new Wrp_KinectFusion();
	Wrp_INuiFusionColorReconstruction winfcr = null;
	Struct_NUI_FUSION_RECONSTRUCTION_PARAMETERS reconstPrm 
		= new Struct_NUI_FUSION_RECONSTRUCTION_PARAMETERS(256f, 512, 384, 512);
	Struct_Matrix4 worldToCameraTransform = new Struct_Matrix4(1f, 0f, 0f, 0f,
																0f, 1f, 0f, 0f,
																0f, 0f, 1f, 0f,
																0f, 0f, 0f, 1f);
	long[] depthFloatFrame = {0}, smoothDepthFloatFrame = {0},
			fusionColorFrame = {0}, pointCloudFrame = {0}, surfaceFrame = {0};
	long[] reconst = {0}, mesh = {0};
	int[] fusionColor;
	int[] surface;
	int errorCount = 0, resetCount = 0;
	final String MESH_FILE = "mesh.ply";
	
	final int IMG_WIDTH = 640;
	final int IMG_HEIGHT = 480;
	final int BORDER_WIDTH = 4;
	final int TITLE_HEIGHT = 30;
	final int PANEL_HEIGHT = 40;
	JButton btnSave, btnReset, btnUp, btnDown, btnExit;
	boolean blSave = false, blReset = false, blUp = false, blDown = false, blExit = false; 
	JProgressBar PrgBar;
	JLabel lblPrg;
	JPanel pnl;
	ImageCanvas imgcvs;
	BufferedImage bfimgOff = new BufferedImage(IMG_WIDTH, IMG_HEIGHT, BufferedImage.TYPE_INT_RGB);
	Graphics grphOff = bfimgOff.getGraphics();
	boolean blPaint = false;
	
	public static void main(String[] args) {
		TestKINECT104 testkinect = new TestKINECT104();
	}
	
	TestKINECT104() {
		this.setTitle("KINECT_TEST104");
		this.setSize(IMG_WIDTH + BORDER_WIDTH * 2, IMG_HEIGHT + TITLE_HEIGHT + PANEL_HEIGHT + BORDER_WIDTH);
		this.setLocationRelativeTo(null);
		this.setLayout(null);
		this.setResizable(false);
		
		imgcvs = new ImageCanvas();
		imgcvs.setBounds(0, 0, IMG_WIDTH, IMG_HEIGHT);
		getContentPane().add(imgcvs);
		
		pnl = new JPanel();
		pnl.setLayout(new FlowLayout(FlowLayout.RIGHT, 10, 5));
		pnl.setBounds(0, IMG_HEIGHT, IMG_WIDTH, PANEL_HEIGHT);
		
		PrgBar = new JProgressBar();
		pnl.add(PrgBar);
		
		lblPrg = new JLabel();
		pnl.add(lblPrg);
		
		btnSave = new JButton("Save Mesh File");
		btnSave.addActionListener(this);
		pnl.add(btnSave);

		btnReset = new JButton(" Reset ");
		btnReset.addActionListener(this);
		pnl.add(btnReset);

		btnUp = new JButton("  Up  ");
		btnUp.addActionListener(this);
		pnl.add(btnUp);
		
		btnDown = new JButton(" Down ");
		btnDown.addActionListener(this);
		pnl.add(btnDown);
		
		btnExit = new JButton(" Exit ");
		btnExit.addActionListener(this);
		pnl.add(btnExit);
		
		getContentPane().add(pnl);
		
		this.addWindowListener(new WindowAdapter() {
			public void windowClosing(WindowEvent ev) {
				System.exit(0);
			}
		});

		this.setVisible(true);
		
		Thread thread = new Thread(this);
		thread.start();
	}
	
	public void actionPerformed(ActionEvent e) {
		JButton btn = (JButton)e.getSource();
		if (btn == btnSave) {
			blSave = true;
		}
		if (btn == btnReset) {
			blReset = true;
		}
		if (btn == btnUp) {
			blUp = true;
		}
		else if (btn == btnDown) {
			blDown = true;
		}
		else if (btn == btnExit) {
			blExit = true;
		}
	}

	class ImageCanvas extends Canvas {
		public void paint(Graphics g){
			if (blPaint == true) {
				g.drawImage(bfimgOff, 0, 0, this);
				blPaint = false;
			}
		}
		
		public void update(Graphics g){
			paint(g);
		}
	}

	public void run() {
		try {
			initialize();
			drawcvs();
		}
		catch (Exception_KINECT e) {
			System.out.println("***** Exception_KINECT: " + e.getMessage());
		}
		catch (Exception_Win64 e) {
			System.out.println("***** Exception_Win64: " + e.getMessage());
		}
		catch (Exception_JInvoker e) {
			System.out.println("***** Exception_JInvoker: " + e.getMessage());
		}
		finally {
			System.out.println("finally start ");
			try {
				terminate();
			}
			catch (Exception_KINECT e) {
				System.out.println("***** Exception_KINECT: " + e.getMessage());
			}
			catch (Exception_Win64 e) {
				System.out.println("***** Exception_Win64: " + e.getMessage());
			}
			System.out.println("finally end ");
		}
		System.exit(0);
	}
	
	void initialize() throws Exception_KINECT, Exception_Win64 {
		// create Kinect instance
		int[] count = {0};
		wk.NuiGetSensorCount(count);
		System.out.println("sensor count: " + Integer.toString(count[0]));
		
		wk.NuiCreateSensorByIndex(0, kinect);
		System.out.println("instance#0: " + Long.toHexString(kinect[0]));
		
		wins = new Wrp_INuiSensor(kinect[0]);
		int sts = wins.NuiStatus();
		System.out.println("status: " + Integer.toHexString(sts));
		
		// initialize
		wins.NuiInitialize(H_NuiApi.NUI_INITIALIZE_FLAG_USES_COLOR
				| H_NuiApi.NUI_INITIALIZE_FLAG_USES_DEPTH_AND_PLAYER_INDEX);
		
		// open color stream
		hColorEvent = wkrnl.CreateEvent(null, true, false, null);
		wins.NuiImageStreamOpen(H_NuiImageCamera.NUI_IMAGE_TYPE_COLOR,
				H_NuiImageCamera.NUI_IMAGE_RESOLUTION_640x480, 0, 2, hColorEvent, hColor);
		System.out.println("color stream: " + Long.toHexString(hColor[0]));
		
		// open depth stream
		hDepthEvent = wkrnl.CreateEvent(null, true, false, null);
		wins.NuiImageStreamOpen(H_NuiImageCamera.NUI_IMAGE_TYPE_DEPTH_AND_PLAYER_INDEX,
				H_NuiImageCamera.NUI_IMAGE_RESOLUTION_640x480, 0, 2, hDepthEvent, hDepth);
		System.out.println("depth stream: " + Long.toHexString(hDepth[0]));

		// near mode
		//wins.NuiImageStreamSetImageFrameFlags(hDepth[0],
		//		H_NuiImageCamera.NUI_IMAGE_STREAM_FLAG_ENABLE_NEAR_MODE);
		
		// get width and height
		wk.NuiImageResolutionToSize(H_NuiImageCamera.NUI_IMAGE_RESOLUTION_640x480, width, height);
		System.out.println("camera size: " + Integer.toString(width[0]) + "x"
				+ Integer.toString(height[0]));
		
		color = new int[width[0] * height[0]];
		depth = new int[width[0] * height[0]];
		colorPoint = new int[width[0] * height[0] * 2]; // x, y
		fusionColor = new int[width[0] * height[0]];
		surface = new int[width[0] * height[0]];
		
		// get coodinate mapper
		wins.NuiGetCoordinateMapper(mapper);
		System.out.println("CoordinateMapper: " + Long.toHexString(mapper[0]));
		wmap = new Wrp_INuiCoordinateMapper(mapper[0]);
		
		// available memory size
		int[] pMemoryKB = {0};
		wkf.NuiFusionGetDeviceInfo(
				H_NuiKinectFusionColorVolume.NUI_FUSION_RECONSTRUCTION_PROCESSOR_TYPE_AMP,
				-1, null, 0, null, 0, pMemoryKB);
		System.out.println("available memory size(KB): " + Long.toString(pMemoryKB[0]));
		
		// create fusion instance
		wkf.NuiFusionCreateColorReconstruction(reconstPrm,
				H_NuiKinectFusionColorVolume.NUI_FUSION_RECONSTRUCTION_PROCESSOR_TYPE_AMP,
				-1, worldToCameraTransform, reconst);
		System.out.println("colorReconstruction: " + Long.toHexString(reconst[0]));
		winfcr = new Wrp_INuiFusionColorReconstruction(reconst[0]);
		
		// create depth image frame for fusion
		wkf.NuiFusionCreateImageFrame(H_NuiKinectFusionColorVolume.NUI_FUSION_IMAGE_TYPE_FLOAT,
				width[0], height[0], null, depthFloatFrame);
		System.out.println("depthFloatFrame: " + Long.toString(depthFloatFrame[0]));
		
		// create smooth depth image frame for fusion
		wkf.NuiFusionCreateImageFrame(H_NuiKinectFusionColorVolume.NUI_FUSION_IMAGE_TYPE_FLOAT,
				width[0], height[0], null, smoothDepthFloatFrame);
		System.out.println("smoothDepthFloatFrame: " + Long.toString(smoothDepthFloatFrame[0]));
		
		// create color image frame for fusion
		wkf.NuiFusionCreateImageFrame(H_NuiKinectFusionColorVolume.NUI_FUSION_IMAGE_TYPE_COLOR,
				width[0], height[0], null, fusionColorFrame);
		System.out.println("fusionColorFrame: " + Long.toString(fusionColorFrame[0]));
		
		// create point cloud image frame for fusion
		wkf.NuiFusionCreateImageFrame(H_NuiKinectFusionColorVolume.NUI_FUSION_IMAGE_TYPE_POINT_CLOUD,
				width[0], height[0], null, pointCloudFrame);
		System.out.println("pointCloudFrame: " + Long.toString(pointCloudFrame[0]));
		
		// create surface image frame for fusion
		wkf.NuiFusionCreateImageFrame(H_NuiKinectFusionColorVolume.NUI_FUSION_IMAGE_TYPE_COLOR,
				width[0], height[0], null, surfaceFrame);
		System.out.println("surfaceFrame: " + Long.toString(surfaceFrame[0]));

		hEvents = new long[]{hColorEvent, hDepthEvent};
	}
	
	void drawcvs() throws Exception_KINECT, Exception_Win64, Exception_JInvoker {
		
		for (;;) {
			
			setSurfaceImage();
			
			blPaint = true;
			imgcvs.repaint();
			
			if (blSave) {
				saveMeshFile();
			}
			if (blReset) {
				reset();
			}
			if (blUp) {
				setAngle(+5);
			}
			if(blDown) {	
				setAngle(-5);
			}
			if (blExit) {
				break;
			}
		}
	}

	void setSurfaceImage() throws Exception_KINECT, Exception_Win64, Exception_JInvoker {
		wkrnl.ResetEvent(hColorEvent);
		wkrnl.ResetEvent(hDepthEvent);
		wkrnl.WaitForMultipleObjects(hEvents.length, hEvents, true, H_WinBase.INFINITE);

		// get color frame
		Struct_NUI_IMAGE_FRAME colorFrame = new Struct_NUI_IMAGE_FRAME();
		wins.NuiImageStreamGetNextFrame(hColor[0], 0, colorFrame);
		
		Wrp_INuiFrameTexture winftColor = new Wrp_INuiFrameTexture(colorFrame.pFrameTexture);
		
		Struct_NUI_LOCKED_RECT colorRect = new Struct_NUI_LOCKED_RECT();
		winftColor.LockRect(0, colorRect, null, 0);

		System.out.println("color data adrs: " + Long.toHexString(colorRect.pBits));
		
		// copy color data
		wjinv.movePtrToIntArray(color, colorRect.pBits);
		
		winftColor.UnlockRect(0);
		wins.NuiImageStreamReleaseFrame(hColor[0], colorFrame);
		
		// get depth frame
		Struct_NUI_IMAGE_FRAME depthFrame = new Struct_NUI_IMAGE_FRAME();
		wins.NuiImageStreamGetNextFrame(hDepth[0], 0, depthFrame);
		
		boolean[] pNearMode = {false};
		long[] ppFrameTexture = {0};
		wins.NuiImageFrameGetDepthImagePixelFrameTexture(hDepth[0], depthFrame, pNearMode,
				ppFrameTexture);
		
		Wrp_INuiFrameTexture winftDepth = new Wrp_INuiFrameTexture(ppFrameTexture[0]);
		
		Struct_NUI_LOCKED_RECT depthRect = new Struct_NUI_LOCKED_RECT();
		winftDepth.LockRect(0, depthRect, null, 0);
		
		System.out.println("depth data adrs: " + Long.toHexString(depthRect.pBits));
		
		// copy depth data
		wjinv.movePtrToIntArray(depth, depthRect.pBits);

		// translate depth to depth float for fusion
		winfcr.DepthToDepthFloatFrame(depthRect.pBits, depthRect.size, depthFloatFrame[0],
				H_NuiKinectFusionDepthProcessor.NUI_FUSION_DEFAULT_MINIMUM_DEPTH,
				H_NuiKinectFusionDepthProcessor.NUI_FUSION_DEFAULT_MAXIMUM_DEPTH, true);
		
		// translate color to color for fusion
		Struct_NUI_FUSION_IMAGE_FRAME snfif = new Struct_NUI_FUSION_IMAGE_FRAME();
		wjinv.movePtrToStruct(snfif, fusionColorFrame[0]);

		Wrp_INuiFrameTexture winftFusionColor = new Wrp_INuiFrameTexture(snfif.pFrameTexture);
		
		Struct_NUI_LOCKED_RECT fusionColorRect = new Struct_NUI_LOCKED_RECT();
		winftFusionColor.LockRect(0, fusionColorRect, null, 0);
		System.out.println("fusion color data adrs: " + Long.toHexString(fusionColorRect.pBits));
		
		wmap.MapDepthFrameToColorFrame(H_NuiImageCamera.NUI_IMAGE_RESOLUTION_640x480,
				depth.length, depthRect.pBits, H_NuiImageCamera.NUI_IMAGE_TYPE_COLOR,
				H_NuiImageCamera.NUI_IMAGE_RESOLUTION_640x480, color.length, colorPoint);
		
		for (int i = 0; i < depth.length; i++) {
			int distance = (depth[i] >> 16) & 0xffff;
			
			int colorX = colorPoint[i * 2];
			int colorY = colorPoint[i * 2 + 1];

			if (distance != 0) {
				if (0 <= colorX && colorX < width[0] && 0 <= colorY && colorY < height[0]) {
					int index = (colorY * width[0]) + colorX;
					fusionColor[i] = color[index];
				}
				else {
					fusionColor[i] = 0xffff0000; // red in case of range over
				}
			}
			else {
				fusionColor[i] = 0xff00ff00; // green in case of no depth data
			}
		}
		
		wjinv.moveIntArrayToPtr(fusionColorRect.pBits, fusionColor);
		
		winftFusionColor.UnlockRect(0);
		winftDepth.UnlockRect(0);
		wins.NuiImageStreamReleaseFrame(hDepth[0], depthFrame);
		
		// smooth depth kernelWidth:1(3x3)/2(5x5)/3(7x7), distanceThreshold:0.01f-0.10f
		winfcr.SmoothDepthFloatFrame(depthFloatFrame[0], smoothDepthFloatFrame[0], 1, 0.04f);
		
		// process depth and color frame
		winfcr.GetCurrentWorldToCameraTransform(worldToCameraTransform);
		int hr = winfcr.ProcessFrame(smoothDepthFloatFrame[0], fusionColorFrame[0],
				(short)H_NuiKinectFusionDepthProcessor.NUI_FUSION_DEFAULT_ALIGN_ITERATION_COUNT,
				(short)H_NuiKinectFusionDepthProcessor.NUI_FUSION_DEFAULT_INTEGRATION_WEIGHT,
				H_NuiKinectFusionDepthProcessor.NUI_FUSION_DEFAULT_COLOR_INTEGRATION_OF_ALL_ANGLES,
				worldToCameraTransform);
		if (H_WinError.FAILED(hr)) {
			errorCount++;
			if (errorCount >= 100) {
				errorCount = 0;
				reset();
			}
		}
		
		// calculate point cloud
		winfcr.CalculatePointCloud(pointCloudFrame[0], surfaceFrame[0], worldToCameraTransform);
		
		// set surface image
		wjinv.movePtrToStruct(snfif, surfaceFrame[0]);
		System.out.println("surfaceFrame pFrameTexture adrs: " + Long.toHexString(snfif.pFrameTexture));

		Wrp_INuiFrameTexture winftSurface = new Wrp_INuiFrameTexture(snfif.pFrameTexture);
		
		Struct_NUI_LOCKED_RECT surfaceRect = new Struct_NUI_LOCKED_RECT();
		winftSurface.LockRect(0, surfaceRect, null, 0);
		
		System.out.println("surface data adrs: " + Long.toHexString(surfaceRect.pBits));
		
		wjinv.movePtrToIntArray(surface, surfaceRect.pBits);
		
		winftSurface.UnlockRect(0);
		
		bfimgOff.setRGB(0, 0, IMG_WIDTH, IMG_HEIGHT, surface, 0, IMG_WIDTH);
		
	}
	
	void saveMeshFile() {
		Wrp_INuiFusionColorMesh winfcm = null;
		long[] pVertices = {0}, pTriangleIndices = {0}, pColors = {0};
		float[] vertices;
		int[] triangleIndices, colors;
		PrintWriter pw = null;
		int curprg = 0, prg = 0;
		
		try {
			// calculate mesh
			winfcr.CalculateMesh(1, mesh);
			System.out.println("color mesh adrs: " + Long.toHexString(mesh[0]));
			winfcm = new Wrp_INuiFusionColorMesh(mesh[0]);
			
			// get counts
			int numVertices = winfcm.VertexCount();
			int numTriangleIndices = winfcm.TriangleVertexIndexCount();
			int numTriangles = numVertices / 3;
			int numColors = winfcm.ColorCount();
			System.out.println("numVertices: " + Integer.toString(numVertices)
					+ " numTriangleIndices: " + Integer.toString(numTriangleIndices)
					+ " numTriangles: " + Integer.toString(numTriangles)
					+ " numColors: " + Integer.toString(numColors));
			
			if (numVertices == 0 || numTriangleIndices == 0 || numVertices % 3 != 0
					|| numTriangleIndices != numVertices || numColors != numVertices) {
				System.out.println("***** vertices number error");
			}
			else {	
				// get vertices
				winfcm.GetVertices(pVertices);
				vertices = new float[numVertices * 3]; // Vector3 {FLOAT x; FLOAT y; FLOAT z;}
				wjinv.movePtrToFloatArray(vertices, pVertices[0]);
				
				// get triangle indices
				winfcm.GetTriangleIndices(pTriangleIndices);
				triangleIndices = new int[numTriangleIndices];
				wjinv.movePtrToIntArray(triangleIndices, pTriangleIndices[0]);
				
				// get colors
				winfcm.GetColors(pColors);
				colors = new int[numColors];
				wjinv.movePtrToIntArray(colors, pColors[0]);
				
				try {
					pw = new PrintWriter(new BufferedWriter(new FileWriter(MESH_FILE)));
					// header
					pw.println("ply");
					pw.println("format ascii 1.0");
					pw.println("comment file created by Kinect Fusion");
						
					// element vertex num, x, y, z, r, g, b
					pw.println(String.format("element vertex %d", numVertices));
					pw.println("property float x");
					pw.println("property float y");
					pw.println("property float z");
					pw.println("property uchar red");
					pw.println("property uchar green");
					pw.println("property uchar blue");
					
					// element face num
					pw.println(String.format("element face %d", numTriangles));
					pw.println("property list uchar int vertex_index");
					pw.println("end_header");
					
					// triangle vertices
					for (int i = 0, idx = 0; i < numVertices; i++, idx += 3) {
						pw.println(String.format("%f %f %f %d %d %d",
								vertices[idx], -vertices[idx + 1], -vertices[idx + 2],
								(colors[i] >> 16) & 0xff, (colors[i] >> 8) & 0xff, colors[i] & 0xff));
						prg = ((i + 1) * 100) / (numVertices + numTriangles);
						if (prg > curprg) {
							PrgBar.setValue(prg);
							lblPrg.setText(Integer.toString(prg) + "%");
							curprg = prg;
						}
					}
	 
					// triangle vertex indices
					for (int i = 0, idx = 0; i < numTriangles; i++, idx += 3) {
						pw.println(String.format("3 %d %d %d", idx, idx + 1, idx + 2));
						prg = ((i + 1 + numVertices) * 100) / (numVertices + numTriangles);
						if (prg > curprg) {
							PrgBar.setValue(prg);
							lblPrg.setText(Integer.toString(prg) + "%");
							curprg = prg;
						}
					}
				} catch (IOException e) {
					System.out.println("***** IOException: " + e.getMessage());
				 }
				finally {
					pw.close();
				}
			}
		}
		catch (Exception_KINECT e) {
			System.out.println("***** Exception_KINECT: " + e.getMessage());
		}
		catch (Exception_JInvoker e) {
			System.out.println("***** Exception_JInvoker: " + e.getMessage());
		}
		finally {
			try {
				if (winfcm != null) winfcm.Release();
				blSave = false;
			}
			catch (Exception_KINECT e) {
				System.out.println("***** Exception_KINECT: " + e.getMessage());
			}
		}
	}
	
	void reset() throws Exception_KINECT {
		resetCount++;
		worldToCameraTransform = new Struct_Matrix4(1f, 0f, 0f, 0f,
													0f, 1f, 0f, 0f,
													0f, 0f, 1f, 0f,
													0f, 0f, 0f, 1f);
		winfcr.ResetReconstruction(worldToCameraTransform, null);
		System.out.println("resetCount: " + Integer.toString(resetCount));
		blReset = false;
	}
	
	void setAngle(int angl) throws Exception_KINECT {
		int[] anglCur = {0};
		int anglSet = 0;
		
		wins.NuiCameraElevationGetAngle(anglCur);
		System.out.println("anglCur: " + Integer.toString(anglCur[0]));
		
		if (angl > 0) {
			anglSet = (anglCur[0] + angl < H_NuiImageCamera.NUI_CAMERA_ELEVATION_MAXIMUM)
					? anglCur[0] + angl : H_NuiImageCamera.NUI_CAMERA_ELEVATION_MAXIMUM;
		}
		else if (angl < 0) {
			anglSet = (anglCur[0] + angl > H_NuiImageCamera.NUI_CAMERA_ELEVATION_MINIMUM)
					? anglCur[0] + angl : H_NuiImageCamera.NUI_CAMERA_ELEVATION_MINIMUM;
		}
		else {
			anglSet = 0;
		}
		wins.NuiCameraElevationSetAngle(anglSet);
		blUp = false;
		blDown = false;
	}
	
	void terminate() throws Exception_KINECT, Exception_Win64 {
		if (winfcr != null) winfcr.Release();
		if (wkf != null) {
			if (depthFloatFrame[0] != 0) wkf.NuiFusionReleaseImageFrame(depthFloatFrame[0]);
			if (smoothDepthFloatFrame[0] != 0) wkf.NuiFusionReleaseImageFrame(smoothDepthFloatFrame[0]);
			if (fusionColorFrame[0] != 0) wkf.NuiFusionReleaseImageFrame(fusionColorFrame[0]);
			if (pointCloudFrame[0] != 0) wkf.NuiFusionReleaseImageFrame(pointCloudFrame[0]);
			if (surfaceFrame[0] != 0) wkf.NuiFusionReleaseImageFrame(surfaceFrame[0]);
		}
		if (kinect[0] != 0) {
			if (wmap != null) {
				wmap.Release();
			}
			if (wins != null) {
				setAngle(0);
				wins.NuiSkeletonTrackingDisable();
				wins.NuiShutdown();
				wins.Release();
			}
		}
		if (wkrnl != null) {
			if (hColorEvent != 0) wkrnl.CloseHandle(hColorEvent);
			if (hDepthEvent != 0) wkrnl.CloseHandle(hDepthEvent);
		}
		if (grphOff != null) grphOff.dispose();
	}
	
}
