package testkinect;

import javax.swing.*;
import javax.swing.event.*;
import java.awt.*;
import java.awt.image.*;
import java.awt.event.*;
import java.io.*;

import utildll.*;

public class TestKINECT105 extends JFrame implements Runnable, ActionListener {
	Wrp_KINECT wk = new Wrp_KINECT();
	Wrp_INuiSensor wins = null;
	Wrp_INuiCoordinateMapper wmap = null;
	Wrp_Kernel32 wkrnl = new Wrp_Kernel32();
	Wrp_JInvoker wjinv = new Wrp_JInvoker();
	long[] kinect = {0};
	long[] hColor = {0};
	long[] hDepth = {0};
	long[] mapper = {0};
	long hColorEvent = 0;
	long hDepthEvent = 0;
	long hSkeletonEvent = 0;
	int[] width = {0};
	int[] height = {0};
	int[] color;
	
	// KinectInteraction data
	Wrp_KinectInteraction wki = new Wrp_KinectInteraction();
	Wrp_INuiInteractionStream winis = null;
	KinectAdapter kadp = new KinectAdapter();
	Struct_pVtbl spvt = new Struct_pVtbl(kadp);
	long[] hInteraction = {0};
	long adapter = 0;
	long hInteractionEvent = 0;
	int[] statusHandLeft, StatusHandRight;
	Struct_POINT[] pointHandLeft, pointHandRight;
	
	final int IMG_WIDTH = 640;
	final int IMG_HEIGHT = 480;
	final int LST_WIDTH = 640;
	final int LST_HEIGHT = 480;
	final int BORDER_WIDTH = 4;
	final int TITLE_HEIGHT = 30;
	final int PANEL_HEIGHT = 40;
	JButton btnUp, btnDown, btnExit;
	boolean blUp = false, blDown = false, blExit = false; 
	JPanel pnl;
	ImageCanvas imgcvs;
	BufferedImage bfimgOff = new BufferedImage(IMG_WIDTH, IMG_HEIGHT, BufferedImage.TYPE_INT_RGB);
	Graphics grphOff = bfimgOff.getGraphics();
	boolean blPaint = false;
	JList listLog; //interaction log
	DefaultListModel model;
	JScrollPane pane;
	
	public static void main(String[] args) {
		TestKINECT105 testkinect = new TestKINECT105();
	}

	TestKINECT105() {
		this.setTitle("KINECT_TEST105");
		this.setSize(IMG_WIDTH + LST_WIDTH + BORDER_WIDTH * 2, IMG_HEIGHT + TITLE_HEIGHT + PANEL_HEIGHT + BORDER_WIDTH);
		this.setLocationRelativeTo(null);
		this.setLayout(null);
		this.setResizable(false);
		
		imgcvs = new ImageCanvas();
		imgcvs.setBounds(0, 0, IMG_WIDTH, IMG_HEIGHT);
		getContentPane().add(imgcvs);
		
		model = new DefaultListModel();
		listLog = new JList(model);
		pane = new JScrollPane();
		pane.getViewport().setView(listLog);
		pane.setBounds(IMG_WIDTH, 0, LST_WIDTH, LST_HEIGHT);
		getContentPane().add(pane);
		
		pnl = new JPanel();
		pnl.setLayout(new FlowLayout(FlowLayout.RIGHT, 10, 5));
		pnl.setBounds(0, IMG_HEIGHT, IMG_WIDTH + LST_WIDTH, PANEL_HEIGHT);
		
		btnUp = new JButton("   Up   ");
		btnUp.addActionListener(this);
		pnl.add(btnUp);
		
		btnDown = new JButton("  Down  ");
		btnDown.addActionListener(this);
		pnl.add(btnDown);
		
		btnExit = new JButton("  Exit  ");
		btnExit.addActionListener(this);
		pnl.add(btnExit);
		
		getContentPane().add(pnl);
		
		this.addWindowListener(new WindowAdapter() {
			public void windowClosing(WindowEvent ev) {
				System.exit(0);
			}
		});

		this.setVisible(true);
		
		Thread thread = new Thread(this);
		thread.start();
	}
	
	public void actionPerformed(ActionEvent e) {
		JButton btn = (JButton)e.getSource();
		if (btn == btnUp) {
			blUp = true;
		}
		else if (btn == btnDown) {
			blDown = true;
		}
		else if (btn == btnExit) {
			blExit = true;
		}
	}

	class ImageCanvas extends Canvas {
		public void paint(Graphics g){
			if (blPaint == true) {
				g.drawImage(bfimgOff, 0, 0, this);
				blPaint = false;
			}
		}
		
		public void update(Graphics g){
			paint(g);
		}
	}

	public void run() {
		try {
			initialize();
			drawcvs();
		}
		catch (Exception_KINECT e) {
			System.out.println("***** Exception_KINECT: " + e.getMessage());
		}
		catch (Exception_Win64 e) {
			System.out.println("***** Exception_Win64: " + e.getMessage());
		}
		catch (Exception_JInvoker e) {
			System.out.println("***** Exception_JInvoker: " + e.getMessage());
		}
		finally {
			System.out.println("finally start ");
			try {
				terminate();
			}
			catch (Exception_KINECT e) {
				System.out.println("***** Exception_KINECT: " + e.getMessage());
			}
			catch (Exception_Win64 e) {
				System.out.println("***** Exception_Win64: " + e.getMessage());
			}
			System.out.println("finally end ");
		}
		System.exit(0);
	}
	
	void initialize() throws Exception_KINECT, Exception_Win64, Exception_JInvoker {
		// create Kinect instance
		int[] count = {0};
		wk.NuiGetSensorCount(count);
		System.out.println("sensor count: " + Integer.toString(count[0]));
		
		wk.NuiCreateSensorByIndex(0, kinect);
		System.out.println("instance#0: " + Long.toHexString(kinect[0]));
		
		wins = new Wrp_INuiSensor(kinect[0]);
		int sts = wins.NuiStatus();
		System.out.println("status: " + Integer.toHexString(sts));
		
		// initialize
		wins.NuiInitialize(H_NuiApi.NUI_INITIALIZE_FLAG_USES_COLOR
				| H_NuiApi.NUI_INITIALIZE_FLAG_USES_DEPTH_AND_PLAYER_INDEX
				| H_NuiApi.NUI_INITIALIZE_FLAG_USES_SKELETON);
		
		// open color stream
		hColorEvent = wkrnl.CreateEvent(null, true, false, null);
		wins.NuiImageStreamOpen(H_NuiImageCamera.NUI_IMAGE_TYPE_COLOR,
				H_NuiImageCamera.NUI_IMAGE_RESOLUTION_640x480, 0, 2, hColorEvent, hColor);
		System.out.println("color stream: " + Long.toHexString(hColor[0]));
		
		// open depth stream
		hDepthEvent = wkrnl.CreateEvent(null, true, false, null);
		wins.NuiImageStreamOpen(H_NuiImageCamera.NUI_IMAGE_TYPE_DEPTH_AND_PLAYER_INDEX,
				H_NuiImageCamera.NUI_IMAGE_RESOLUTION_640x480, 0, 2, hDepthEvent, hDepth);
		System.out.println("depth stream: " + Long.toHexString(hDepth[0]));

		// near mode
		//wins.NuiImageStreamSetImageFrameFlags(hDepth[0],
		//		H_NuiImageCamera.NUI_IMAGE_STREAM_FLAG_ENABLE_NEAR_MODE);
		
		// enable skeleton tracking
		hSkeletonEvent = wkrnl.CreateEvent(null, true, false, null);
		wins.NuiSkeletonTrackingEnable(hSkeletonEvent, 0);
		
		// create interaction stream
		adapter = wjinv.allocStruct(spvt);
		wki.NuiCreateInteractionStream(kinect[0], adapter, hInteraction);
		System.out.println("interaction stream: " + Long.toHexString(hInteraction[0]));
		
		winis = new Wrp_INuiInteractionStream(hInteraction[0]);
		
		hInteractionEvent = wkrnl.CreateEvent(null, true, false, null);
		winis.Enable(hInteractionEvent);
		
		// get width and height
		wk.NuiImageResolutionToSize(H_NuiImageCamera.NUI_IMAGE_RESOLUTION_640x480, width, height);
		System.out.println("camera size: " + Integer.toString(width[0]) + "x"
				+ Integer.toString(height[0]));
		
		color = new int[width[0] * height[0]];
		
		// get coodinate mapper
		wins.NuiGetCoordinateMapper(mapper);
		System.out.println("CoordinateMapper: " + Long.toHexString(mapper[0]));
		wmap = new Wrp_INuiCoordinateMapper(mapper[0]);
		
		// hands status and position
		statusHandLeft = new int[H_NuiSkeleton.NUI_SKELETON_COUNT];
		StatusHandRight = new int[H_NuiSkeleton.NUI_SKELETON_COUNT];
		pointHandLeft = new Struct_POINT[H_NuiSkeleton.NUI_SKELETON_COUNT];
		pointHandRight = new Struct_POINT[H_NuiSkeleton.NUI_SKELETON_COUNT];
	}
	
	void drawcvs() throws Exception_KINECT, Exception_Win64, Exception_JInvoker {
		
		for (;;) {
			
			setColorImage();
			drawSkeleton();
			drawInteraction();
			
			blPaint = true;
			imgcvs.repaint();
			
			if (blUp) {
				setAngle(+5);
			}
			if(blDown) {	
				setAngle(-5);
			}
			if (blExit) {
				break;
			}
		}
	}

	void setColorImage() throws Exception_KINECT, Exception_Win64, Exception_JInvoker {
		int rtn;
		rtn = wkrnl.WaitForSingleObject(hColorEvent, 0);
		if (rtn == H_WinBase.WAIT_OBJECT_0) {
			// get color frame
			Struct_NUI_IMAGE_FRAME colorFrame = new Struct_NUI_IMAGE_FRAME();
			wins.NuiImageStreamGetNextFrame(hColor[0], 0, colorFrame);
			
			Wrp_INuiFrameTexture winftColor = new Wrp_INuiFrameTexture(colorFrame.pFrameTexture);
			
			Struct_NUI_LOCKED_RECT colorRect = new Struct_NUI_LOCKED_RECT();
			winftColor.LockRect(0, colorRect, null, 0);
	
			System.out.println("color data adrs: " + Long.toHexString(colorRect.pBits));
			
			// copy color data
			wjinv.movePtrToIntArray(color, colorRect.pBits);
			
			winftColor.UnlockRect(0);
			wins.NuiImageStreamReleaseFrame(hColor[0], colorFrame);
			
			// set color image
			bfimgOff.setRGB(0, 0, IMG_WIDTH, IMG_HEIGHT, color, 0, IMG_WIDTH);
		}
		
		rtn = wkrnl.WaitForSingleObject(hDepthEvent, 0);
		if (rtn == H_WinBase.WAIT_OBJECT_0) {
			// get depth frame
			Struct_NUI_IMAGE_FRAME depthFrame = new Struct_NUI_IMAGE_FRAME();
			wins.NuiImageStreamGetNextFrame(hDepth[0], 0, depthFrame);
			
			boolean[] pNearMode = {false};
			long[] ppFrameTexture = {0};
			wins.NuiImageFrameGetDepthImagePixelFrameTexture(hDepth[0], depthFrame, pNearMode,
					ppFrameTexture);
			
			Wrp_INuiFrameTexture winftDepth = new Wrp_INuiFrameTexture(ppFrameTexture[0]);
			
			Struct_NUI_LOCKED_RECT depthRect = new Struct_NUI_LOCKED_RECT();
			winftDepth.LockRect(0, depthRect, null, 0);
			
			System.out.println("depth data adrs: " + Long.toHexString(depthRect.pBits));
			
			// depth process for interaction
			winis.ProcessDepth(depthRect.size, depthRect.pBits, depthFrame.liTimeStamp);

			winftDepth.UnlockRect(0);
			wins.NuiImageStreamReleaseFrame(hDepth[0], depthFrame);
		}
	}
	
	void drawSkeleton() throws Exception_KINECT, Exception_Win64 {
		int rtn = wkrnl.WaitForSingleObject(hSkeletonEvent, 0);
		if (rtn == H_WinBase.WAIT_OBJECT_0) {
			// get skeleton frame
			Struct_NUI_SKELETON_FRAME skeletonFrame = new Struct_NUI_SKELETON_FRAME();
			int hr = wins.NuiSkeletonGetNextFrame(0, skeletonFrame);
		
			// clear hands position
			for (int i = 0; i < H_NuiSkeleton.NUI_SKELETON_COUNT; i++) {
				pointHandLeft[i] = null;
				pointHandRight[i] = null;
			}
			
			// draw skeleton
			if (H_WinError.SUCCEEDED(hr)) {
				for (int i = 0; i < H_NuiSkeleton.NUI_SKELETON_COUNT; i++) {
					Struct_NUI_SKELETON_DATA skeletonData = skeletonFrame.SkeletonData[i];
					System.out.println("SkeletonData[" + Integer.toString(i) + "].eTrackingState: "
								+ Integer.toString(skeletonData.eTrackingState));
		
					if (skeletonData.eTrackingState == H_NuiSkeleton.NUI_SKELETON_TRACKED) {
						Struct_POINT[] sp = new Struct_POINT[H_NuiSkeleton.NUI_SKELETON_POSITION_COUNT];
						for (int j = 0; j < H_NuiSkeleton.NUI_SKELETON_POSITION_COUNT; j++) {
							if (skeletonData.eSkeletonPositionTrackingState[j]
									!= H_NuiSkeleton.NUI_SKELETON_POSITION_NOT_TRACKED) {
								sp[j] = drawJoint(skeletonData.SkeletonPositions[j]);
							}
						}
						drawLine(sp);
						
						// save hands position
						pointHandLeft[i] = sp[H_NuiSkeleton.NUI_SKELETON_POSITION_HAND_LEFT];
						pointHandRight[i] = sp[H_NuiSkeleton.NUI_SKELETON_POSITION_HAND_RIGHT];
					}
					else if (skeletonData.eTrackingState == H_NuiSkeleton.NUI_SKELETON_POSITION_ONLY) {
						drawJoint(skeletonData.Position);
					}
				}
				
				// skeleton process for interaction
				Struct_Vector4 reading = new Struct_Vector4();
				wins.NuiAccelerometerGetCurrentReading(reading);
				winis.ProcessSkeleton(H_NuiSkeleton.NUI_SKELETON_COUNT, skeletonFrame.SkeletonData,
						reading, skeletonFrame.liTimeStamp);
			}
		}
	}

	Struct_POINT drawJoint(Struct_Vector4 position) throws Exception_KINECT {
		int[] pColorPoint = {0, 0};
		wmap.MapSkeletonPointToColorPoint(position, H_NuiImageCamera.NUI_IMAGE_TYPE_COLOR,
				H_NuiImageCamera.NUI_IMAGE_RESOLUTION_640x480, pColorPoint);
		
		//System.out.println("MapSkeletonPointToColorPoint: " + Integer.toString(pColorPoint[0])
		//		+ ", " + Integer.toString(pColorPoint[1]));
		
		grphOff.setColor(Color.green);
		grphOff.drawOval(pColorPoint[0], pColorPoint[1], 5, 5);
		
		return new Struct_POINT(pColorPoint[0], pColorPoint[1]);
	}
	
	void drawLine(Struct_POINT[] sp) {
		for (int i = 0; i < H_NuiSkeleton.NUI_SKELETON_POSITION_COUNT; i++) {
			if (sp[i] == null) return;
		}
		
		int[] pointx = new int[5];
		int[] pointy = new int[5];
		grphOff.setColor(Color.green);
		
		pointx[0] = sp[H_NuiSkeleton.NUI_SKELETON_POSITION_HEAD].x;
		pointy[0] = sp[H_NuiSkeleton.NUI_SKELETON_POSITION_HEAD].y;
		pointx[1] = sp[H_NuiSkeleton.NUI_SKELETON_POSITION_SHOULDER_CENTER].x;
		pointy[1] = sp[H_NuiSkeleton.NUI_SKELETON_POSITION_SHOULDER_CENTER].y;
		pointx[2] = sp[H_NuiSkeleton.NUI_SKELETON_POSITION_SPINE].x;
		pointy[2] = sp[H_NuiSkeleton.NUI_SKELETON_POSITION_SPINE].y;
		pointx[3] = sp[H_NuiSkeleton.NUI_SKELETON_POSITION_HIP_CENTER].x;
		pointy[3] = sp[H_NuiSkeleton.NUI_SKELETON_POSITION_HIP_CENTER].y;
		grphOff.drawPolyline(pointx, pointy, 4);
		
		pointx[0] = sp[H_NuiSkeleton.NUI_SKELETON_POSITION_SHOULDER_CENTER].x;
		pointy[0] = sp[H_NuiSkeleton.NUI_SKELETON_POSITION_SHOULDER_CENTER].y;
		pointx[1] = sp[H_NuiSkeleton.NUI_SKELETON_POSITION_SHOULDER_LEFT].x;
		pointy[1] = sp[H_NuiSkeleton.NUI_SKELETON_POSITION_SHOULDER_LEFT].y;
		pointx[2] = sp[H_NuiSkeleton.NUI_SKELETON_POSITION_ELBOW_LEFT].x;
		pointy[2] = sp[H_NuiSkeleton.NUI_SKELETON_POSITION_ELBOW_LEFT].y;
		pointx[3] = sp[H_NuiSkeleton.NUI_SKELETON_POSITION_WRIST_LEFT].x;
		pointy[3] = sp[H_NuiSkeleton.NUI_SKELETON_POSITION_WRIST_LEFT].y;
		pointx[4] = sp[H_NuiSkeleton.NUI_SKELETON_POSITION_HAND_LEFT].x;
		pointy[4] = sp[H_NuiSkeleton.NUI_SKELETON_POSITION_HAND_LEFT].y;
		grphOff.drawPolyline(pointx, pointy, 5);
		
		pointx[0] = sp[H_NuiSkeleton.NUI_SKELETON_POSITION_SHOULDER_CENTER].x;
		pointy[0] = sp[H_NuiSkeleton.NUI_SKELETON_POSITION_SHOULDER_CENTER].y;
		pointx[1] = sp[H_NuiSkeleton.NUI_SKELETON_POSITION_SHOULDER_RIGHT].x;
		pointy[1] = sp[H_NuiSkeleton.NUI_SKELETON_POSITION_SHOULDER_RIGHT].y;
		pointx[2] = sp[H_NuiSkeleton.NUI_SKELETON_POSITION_ELBOW_RIGHT].x;
		pointy[2] = sp[H_NuiSkeleton.NUI_SKELETON_POSITION_ELBOW_RIGHT].y;
		pointx[3] = sp[H_NuiSkeleton.NUI_SKELETON_POSITION_WRIST_RIGHT].x;
		pointy[3] = sp[H_NuiSkeleton.NUI_SKELETON_POSITION_WRIST_RIGHT].y;
		pointx[4] = sp[H_NuiSkeleton.NUI_SKELETON_POSITION_HAND_RIGHT].x;
		pointy[4] = sp[H_NuiSkeleton.NUI_SKELETON_POSITION_HAND_RIGHT].y;
		grphOff.drawPolyline(pointx, pointy, 5);
		
		pointx[0] = sp[H_NuiSkeleton.NUI_SKELETON_POSITION_HIP_CENTER].x;
		pointy[0] = sp[H_NuiSkeleton.NUI_SKELETON_POSITION_HIP_CENTER].y;
		pointx[1] = sp[H_NuiSkeleton.NUI_SKELETON_POSITION_HIP_LEFT].x;
		pointy[1] = sp[H_NuiSkeleton.NUI_SKELETON_POSITION_HIP_LEFT].y;
		pointx[2] = sp[H_NuiSkeleton.NUI_SKELETON_POSITION_KNEE_LEFT].x;
		pointy[2] = sp[H_NuiSkeleton.NUI_SKELETON_POSITION_KNEE_LEFT].y;
		pointx[3] = sp[H_NuiSkeleton.NUI_SKELETON_POSITION_ANKLE_LEFT].x;
		pointy[3] = sp[H_NuiSkeleton.NUI_SKELETON_POSITION_ANKLE_LEFT].y;
		pointx[4] = sp[H_NuiSkeleton.NUI_SKELETON_POSITION_FOOT_LEFT].x;
		pointy[4] = sp[H_NuiSkeleton.NUI_SKELETON_POSITION_FOOT_LEFT].y;
		grphOff.drawPolyline(pointx, pointy, 5);
		
		pointx[0] = sp[H_NuiSkeleton.NUI_SKELETON_POSITION_HIP_CENTER].x;
		pointy[0] = sp[H_NuiSkeleton.NUI_SKELETON_POSITION_HIP_CENTER].y;
		pointx[1] = sp[H_NuiSkeleton.NUI_SKELETON_POSITION_HIP_RIGHT].x;
		pointy[1] = sp[H_NuiSkeleton.NUI_SKELETON_POSITION_HIP_RIGHT].y;
		pointx[2] = sp[H_NuiSkeleton.NUI_SKELETON_POSITION_KNEE_RIGHT].x;
		pointy[2] = sp[H_NuiSkeleton.NUI_SKELETON_POSITION_KNEE_RIGHT].y;
		pointx[3] = sp[H_NuiSkeleton.NUI_SKELETON_POSITION_ANKLE_RIGHT].x;
		pointy[3] = sp[H_NuiSkeleton.NUI_SKELETON_POSITION_ANKLE_RIGHT].y;
		pointx[4] = sp[H_NuiSkeleton.NUI_SKELETON_POSITION_FOOT_RIGHT].x;
		pointy[4] = sp[H_NuiSkeleton.NUI_SKELETON_POSITION_FOOT_RIGHT].y;
		grphOff.drawPolyline(pointx, pointy, 5);
	}

	void drawInteraction() throws Exception_KINECT, Exception_Win64 {
		int rtn = wkrnl.WaitForSingleObject(hInteractionEvent, 0);
		if (rtn == H_WinBase.WAIT_OBJECT_0) {
			Struct_NUI_INTERACTION_FRAME interactionFrame = new Struct_NUI_INTERACTION_FRAME(); 
			int hr = winis.GetNextFrame(0, interactionFrame);
			
			if (H_WinError.SUCCEEDED(hr)) {
				for (int i = 0; i < H_NuiSkeleton.NUI_SKELETON_COUNT; i++) {
					Struct_NUI_USER_INFO userInfo = interactionFrame.UserInfos[i];
					if (userInfo.SkeletonTrackingId != 0) {
						for (int j = 0; j < H_KinectInteraction.NUI_USER_HANDPOINTER_COUNT; j++) {
							Struct_NUI_HANDPOINTER_INFO handInfo = userInfo.HandPointerInfos[j];
							//fileLog(toStrHandInfo(handInfo));
							if ((handInfo.State & H_KinectInteraction.NUI_HANDPOINTER_STATE_TRACKED) != 0) {
								if (handInfo.HandEventType == H_KinectInteraction.NUI_HAND_EVENT_TYPE_GRIP) {
									if (handInfo.HandType == H_KinectInteraction.NUI_HAND_TYPE_LEFT) {
										statusHandLeft[i] = H_KinectInteraction.NUI_HAND_EVENT_TYPE_GRIP;
									}
									else if (handInfo.HandType == H_KinectInteraction.NUI_HAND_TYPE_RIGHT) {
										StatusHandRight[i] =  H_KinectInteraction.NUI_HAND_EVENT_TYPE_GRIP;
									}								
									appendLog(toStrHandInfo(handInfo));
								}
								else if (handInfo.HandEventType == H_KinectInteraction.NUI_HAND_EVENT_TYPE_GRIPRELEASE) {
									if (handInfo.HandType == H_KinectInteraction.NUI_HAND_TYPE_LEFT) {
										statusHandLeft[i] = H_KinectInteraction.NUI_HAND_EVENT_TYPE_GRIPRELEASE;
									}
									else if (handInfo.HandType == H_KinectInteraction.NUI_HAND_TYPE_RIGHT) {
										StatusHandRight[i] =  H_KinectInteraction.NUI_HAND_EVENT_TYPE_GRIPRELEASE;
									}								
									appendLog(toStrHandInfo(handInfo));
								}
								
								// draw hands 
								Struct_POINT pointHand = null;
								int statusHand = H_KinectInteraction.NUI_HAND_EVENT_TYPE_NONE;
								if (handInfo.HandType == H_KinectInteraction.NUI_HAND_TYPE_LEFT) {
									pointHand = pointHandLeft[i];
									statusHand = statusHandLeft[i];
								}
								else if (handInfo.HandType == H_KinectInteraction.NUI_HAND_TYPE_RIGHT) {
									pointHand = pointHandRight[i];
									statusHand = StatusHandRight[i];
								}								
								
								if (pointHand != null) {
									if (statusHand == H_KinectInteraction.NUI_HAND_EVENT_TYPE_GRIP) {
										if ((handInfo.State & H_KinectInteraction.NUI_HANDPOINTER_STATE_PRESSED) != 0) {
											// grip and pressed
											grphOff.setColor(Color.red);
											grphOff.fillRect(pointHand.x - 20, pointHand.y - 20, 40, 40);
										}
										else {
											// grip
											grphOff.setColor(Color.yellow);
											grphOff.fillRect(pointHand.x - 20, pointHand.y - 20, 40, 40);
										}
									}
									else if (statusHand == H_KinectInteraction.NUI_HAND_EVENT_TYPE_GRIPRELEASE) {
										if ((handInfo.State & H_KinectInteraction.NUI_HANDPOINTER_STATE_PRESSED) != 0) {
											// release and pressed
											grphOff.setColor(Color.red);
											grphOff.fillOval(pointHand.x, pointHand.y, 40, 40);

										}
										else {
											// release
											grphOff.setColor(Color.yellow);
											grphOff.fillOval(pointHand.x, pointHand.y, 40, 40);
										}
									}
								}
							}
							else { // not tracked
								if (handInfo.HandType == H_KinectInteraction.NUI_HAND_TYPE_LEFT) {
									statusHandLeft[i] = H_KinectInteraction.NUI_HAND_EVENT_TYPE_NONE;
								}
								else if (handInfo.HandType == H_KinectInteraction.NUI_HAND_TYPE_RIGHT) {
									StatusHandRight[i] =  H_KinectInteraction.NUI_HAND_EVENT_TYPE_NONE;
								}								
							}
						}
						//fileLog("-----------");
					}
					else {
						statusHandLeft[i] = H_KinectInteraction.NUI_HAND_EVENT_TYPE_NONE;
						StatusHandRight[i] =  H_KinectInteraction.NUI_HAND_EVENT_TYPE_NONE;
					}
				}
			}
		}
	}
	
	String toStrHandInfo(Struct_NUI_HANDPOINTER_INFO handInfo) {
		StringBuilder sb = new StringBuilder();
		// hand type
		if (handInfo.HandType == H_KinectInteraction.NUI_HAND_TYPE_NONE) {
			sb.append("none");
		}
		else if (handInfo.HandType == H_KinectInteraction.NUI_HAND_TYPE_LEFT) {
			sb.append("left");
		}
		else if (handInfo.HandType == H_KinectInteraction.NUI_HAND_TYPE_RIGHT) {
			sb.append("right");
		}
		// event type
		if (handInfo.HandEventType == H_KinectInteraction.NUI_HAND_EVENT_TYPE_NONE) {
			sb.append(" none-event");
		}
		else if (handInfo.HandEventType == H_KinectInteraction.NUI_HAND_EVENT_TYPE_GRIP) {
			sb.append(" grip");
		}
		else if (handInfo.HandEventType == H_KinectInteraction.NUI_HAND_EVENT_TYPE_GRIPRELEASE) {
			sb.append(" grip-release");
		}
		// hand state
		sb.append(" state: ");
		sb.append(Integer.toHexString(handInfo.State));
		sb.append(" TRACKED=");
		sb.append(Integer.toHexString(handInfo.State & 0x1));
		sb.append(" ACTIVE=");
		sb.append(Integer.toHexString((handInfo.State >> 1) & 0x1));
		sb.append(" INTERACTIVE=");
		sb.append(Integer.toHexString((handInfo.State >> 2) & 0x1));
		sb.append(" PRESSED=");
		sb.append(Integer.toHexString((handInfo.State >> 3) & 0x1));
		sb.append(" PRIMARY_FOR_USER=");
		sb.append(Integer.toHexString((handInfo.State >> 4) & 0x1));
		// X,Y
		sb.append(" X,Y: ");
		sb.append(Float.toString(handInfo.X));
		sb.append(",");
		sb.append(Float.toString(handInfo.Y));
		// RawX,RawY,RawZ
		sb.append(" RawX,RawY,RawZ: ");
		sb.append(Float.toString(handInfo.RawX));
		sb.append(",");
		sb.append(Float.toString(handInfo.RawY));
		sb.append(",");
		sb.append(Float.toString(handInfo.RawZ));
		
		return sb.toString();
	}
	
	void appendLog(String log) {
		model.addElement(log);
		int lastIndex = model.getSize() - 1;
		if (lastIndex >= 0) {
			listLog.ensureIndexIsVisible(lastIndex);
		}
	}

	void fileLog(String str) {
        File file = new File("log.txt");
        PrintWriter pw = null;
 
        try {
            pw = new PrintWriter(
                    new OutputStreamWriter(
                            new FileOutputStream(file, true)));
 
            pw.println(str);
        } catch (IOException e) {
            System.out.println(e);
        } finally {
            if (pw != null) {
                pw.close();
            }
        }
	}
	
	void setAngle(int angl) throws Exception_KINECT {
		int[] anglCur = {0};
		int anglSet = 0;
		
		wins.NuiCameraElevationGetAngle(anglCur);
		System.out.println("anglCur: " + Integer.toString(anglCur[0]));
		
		if (angl > 0) {
			anglSet = (anglCur[0] + angl < H_NuiImageCamera.NUI_CAMERA_ELEVATION_MAXIMUM)
					? anglCur[0] + angl : H_NuiImageCamera.NUI_CAMERA_ELEVATION_MAXIMUM;
		}
		else if (angl < 0) {
			anglSet = (anglCur[0] + angl > H_NuiImageCamera.NUI_CAMERA_ELEVATION_MINIMUM)
					? anglCur[0] + angl : H_NuiImageCamera.NUI_CAMERA_ELEVATION_MINIMUM;
		}
		else {
			anglSet = 0;
		}
		wins.NuiCameraElevationSetAngle(anglSet);
		blUp = false;
		blDown = false;
	}
	
	void terminate() throws Exception_KINECT, Exception_Win64 {
		
		if (kinect[0] != 0) {
			if (wmap != null) {
				wmap.Release();
			}
			if (winis != null) {
				winis.Disable();
				winis.Release();
			}
			if (wins != null) {
				setAngle(0);
				wins.NuiSkeletonTrackingDisable();
				wins.NuiShutdown();
				wins.Release();
			}
		}
		if (wkrnl != null) {
			if (hColorEvent != 0) wkrnl.CloseHandle(hColorEvent);
			if (hDepthEvent != 0) wkrnl.CloseHandle(hDepthEvent);
			if (hSkeletonEvent != 0) wkrnl.CloseHandle(hSkeletonEvent);
			if (hInteractionEvent != 0) wkrnl.CloseHandle(hInteractionEvent);
		}
		if (grphOff != null) grphOff.dispose();
	}
	
	class KinectAdapter {
		int QueryInterface(EventArg_QueryInterface aqi){
			System.out.println("[KinectAdapter] QueryInterface called ");
			appendLog("[KinectAdapter] QueryInterface called ");
			return H_WinError.E_NOINTERFACE;
		}
		
		int AddRef(EventArg_AddRef aar){
			System.out.println("[KinectAdapter] AddRef called ");
			appendLog("[KinectAdapter] AddRef called ");
			return 1;
		}
		
		int Release(EventArg_Release ar){
			System.out.println("[KinectAdapter] Release called ");
			appendLog("[KinectAdapter] Release called ");
			return 1;
		}
		
		int GetInteractionInfoAtLocation(EventArg_GetInteractionInfoAtLocation agiial){
			System.out.println("[KinectAdapter] GetInteractionInfoAtLocation called ");
			String val = "IsGripTarget: " + Boolean.toString(agiial.pInteractionInfo.IsGripTarget)
					+ " IsPressTarget: " + Boolean.toString(agiial.pInteractionInfo.IsPressTarget)
					+ " PressTargetControlId: " + Integer.toString(agiial.pInteractionInfo.PressTargetControlId)
					+ " PressAttractionPointX: " + Float.toString(agiial.pInteractionInfo.PressAttractionPointX)
					+ " PressAttractionPointY: " + Float.toString(agiial.pInteractionInfo.PressAttractionPointY);
			System.out.println(val);
			//apendLog(val);
			agiial.pInteractionInfo.IsGripTarget = false;
			agiial.pInteractionInfo.IsPressTarget = false;
			agiial.pInteractionInfo.PressTargetControlId = 0;
			agiial.pInteractionInfo.PressAttractionPointX = 0f;
			agiial.pInteractionInfo.PressAttractionPointY = 0f;
			return H_WinError.S_OK;
		}
	}

	class Struct_pVtbl{
		Struct_Vtbl pVtbl;
		
		Struct_pVtbl(Object _Instance){
			this.pVtbl = new Struct_Vtbl(_Instance);
		}
	}
	
	class Struct_Vtbl{
		@_FuncPtr Method_def pQueryInterface;
		@_FuncPtr Method_def pAddRef;
		@_FuncPtr Method_def pRelease;
		@_FuncPtr Method_def pGetInteractionInfoAtLocation;
		
		Struct_Vtbl(Object _Instance) {
			this.pQueryInterface = new Method_def("QueryInterface", Func.RTNTYPE_INT,
					new EventArg_QueryInterface(null, null), _Instance);
			this.pAddRef = new Method_def("AddRef", Func.RTNTYPE_INT,
					new EventArg_AddRef(), _Instance);
			this.pRelease = new Method_def("Release", Func.RTNTYPE_INT,
					new EventArg_Release(), _Instance);
			this.pGetInteractionInfoAtLocation = new Method_def("GetInteractionInfoAtLocation", Func.RTNTYPE_INT,
					new EventArg_GetInteractionInfoAtLocation(0, 0, 0f, 0f, null), _Instance);
		}
	}

	class Method_def {
		private String _FuncName;
		private int _ReturnType;
		private Object _FuncArg;
		private Object _Instance;
		
		Method_def(String _FuncName, int _ReturnType, Object _FuncArg, Object _Instance){
			this._FuncName = _FuncName;
			this._ReturnType = _ReturnType;
			this._FuncArg = _FuncArg;
			this._Instance = _Instance;
		}
	}

	class EventArg_property {
		private long _This;
	
		long getThis() {return _This;}
	}

	/*
	HRESULT QueryInterface( 
		REFIID iid, 
		void **ppvObject
	);
	 */
	class EventArg_QueryInterface extends EventArg_property {
		private @_In Struct_IID iid;
		private @_Out long[] ppvObject;
		
		EventArg_QueryInterface(Struct_IID iid, long[] ppvObject){
			this.iid = iid;
			this.ppvObject = ppvObject;
		}
	}
	
	/*
	ULONG AddRef( void );
	 */
	class EventArg_AddRef extends EventArg_property {
	}
	
	/*
	ULONG Release( void );
	 */
	class EventArg_Release extends EventArg_property {
	}

	/*
	HRESULT GetInteractionInfoAtLocation( 
		DWORD skeletonTrackingId, 
		NUI_HAND_TYPE handType,
		FLOAT x,
		FLOAT y,
		NUI_INTERACTION_INFO *pInteractionInfo
	);
	 */
	class EventArg_GetInteractionInfoAtLocation extends EventArg_property {
		private int skeletonTrackingId;
		private int handType;
		private float x;
		private float y;
		private @_Out Struct_NUI_INTERACTION_INFO pInteractionInfo;
		
		EventArg_GetInteractionInfoAtLocation(int skeletonTrackingId, int handType, float x, float y,
				Struct_NUI_INTERACTION_INFO pInteractionInfo) {
			this.skeletonTrackingId = skeletonTrackingId;
			this.handType = handType;
			this.x = x;
			this.y = y;
			this.pInteractionInfo = pInteractionInfo;
		}
	}
	
}
